import re


# custom exceptions
class AWSEMJobErrorException(Exception):
    """There is an error from a worklow run on the EC2 AWSEM instance."""
    pass


class AWSEMErrorHandler(object):

    def __init__(self):
        self.ErrorList = self._ErrorList  # initial error list, custom errors can be added

    class AWSEMError(object):
        def __init__(self, error_type, pattern_in_log, multiline=False):
            self.error_type = error_type
            if multiline:
                self.pattern_in_log = re.compile(pattern_in_log, re.MULTILINE)
            else:
                self.pattern_in_log = pattern_in_log

    def add_custom_errors(self, custom_err_list):
        """add custom errors to ErrorList.
        custom_err_list is a list of dictionaries w/ keys 'error_type', 'pattern', 'multiline'"""
        for err in custom_err_list:
            self.ErrorList.append(self.AWSEMError(err['error_type'], err['pattern'], err.get('multiline', False)))

    @property
    def _ErrorList(self):
        """add any specific error types with recognizable strings or patterns here.
        the order is important. The earlier ones are checked first and if there is a match,
        the later ones will not be checked."""
        return [
            # input download failure due to not enough disk space
            self.AWSEMError('Not enough space for input files', 'download failed: .+ No space left on device'),
            # Docker pull failure due to not enough root disk space
            self.AWSEMError('No space for docker', 'failed to register layer.+no space left on device'),
            # not enough disk space
            self.AWSEMError('Not enough space', '.+No space left on device'),
            # CWL missing input error
            self.AWSEMError('CWL missing input', 'Missing required input parameter\n.+\n', True),
            # Bucket access error
            self.AWSEMError('Bucket/file access denied', 'when calling the ListObjectsV2 operation: Access Denied')
        ]

    def parse_log(self, log):
        # for ex in self.AWSEMErrorExceptionList:
        for ex in self.ErrorList:
            res = re.search(ex.pattern_in_log, log)
            if res:
                match = res.string[res.regs[0][0]:res.regs[0][1]]
                match = re.sub('\n', ' ', match)  # \n not recognized and subsequent content is dropped from Exception
                match = re.sub(' +', ' ', match)
                msg = "%s: %s" % (ex.error_type, match)
                return AWSEMJobErrorException(msg)
        return

    @property
    def general_awsem_check_log_msg_template(self):
        return "check log using tibanna log --job-id=%s [--sfn=stepfunction]"

    def general_awsem_check_log_msg(self, job_id):
        return self.general_awsem_check_log_msg_template % job_id

    @property
    def general_awsem_error_msg_template(self):
        return "Job encountered an error " + self.general_awsem_check_log_msg_template

    def general_awsem_error_msg(self, job_id):
        return self.general_awsem_error_msg_template % job_id


class StillRunningException(Exception):
    """EC2 AWSEM instance is still running (job not complete)"""
    pass


class EC2StartingException(Exception):
    """EC2 AWSEM instance is still starting (job not complete)"""
    pass


class DependencyStillRunningException(Exception):
    pass


class DependencyFailedException(Exception):
    pass


class EC2LaunchException(Exception):
    pass


class EC2UnintendedTerminationException(Exception):
    pass


class EC2IdleException(Exception):
    pass


class EC2InstanceLimitException(Exception):
    pass


class EC2InstanceLimitWaitException(Exception):
    pass


class MissingFieldInInputJsonException(Exception):
    pass


class MalFormattedInputJsonException(Exception):
    pass


class MalFormattedRunJsonException(Exception):
    pass


class MalFormattedPostRunJsonException(Exception):
    pass


class MetricRetrievalException(Exception):
    pass


class PricingRetrievalException(Exception):
    pass


class UnsupportedCWLVersionException(Exception):
    def __init__(self, message=None):
        if not message:
            message = "CWL draft3 is no longer supported. Please switched to v1"
        super().__init__(message)


class JobAbortedException(Exception):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['awsf3', 'tibanna', 'tibanna.lambdas']

package_data = \
{'': ['*']}

install_requires = \
['Benchmark-4dn>=0.5.8,<0.6.0',
 'boto3>=1.9.0,<2.0.0',
 'botocore>=1.12.1,<2.0.0',
 'python-lambda-4dn==0.12.3',
 'requests>=2.22.0,<3.0.0',
 'tomlkit>=0.10.0,<0.11.0',
 'urllib3>=1.24,<2.0']

entry_points = \
{'console_scripts': ['awsf3 = awsf3.__main__:main',
                     'tibanna = tibanna.__main__:main']}

setup_kwargs = {
    'name': 'tibanna',
    'version': '2.2.0',
    'description': 'Tibanna runs portable pipelines (in CWL/WDL) on the AWS Cloud.',
    'long_description': "# Tibanna\n\n[![Python 3.8](https://img.shields.io/badge/python-3.8-blue.svg)](https://www.python.org/downloads/release/python-380/) [![Build Status](https://travis-ci.org/4dn-dcic/tibanna.svg?branch=master)](https://travis-ci.org/4dn-dcic/tibanna) [![Code Quality](https://api.codacy.com/project/badge/Grade/d2946b5bc0704e5c9a4893426a7e0314)](https://www.codacy.com/app/4dn/tibanna?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=4dn-dcic/tibanna&amp;utm_campaign=Badge_Grade) [![Test Coverage](https://api.codacy.com/project/badge/Coverage/d2946b5bc0704e5c9a4893426a7e0314)](https://www.codacy.com/app/4dn/tibanna?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=4dn-dcic/tibanna&amp;utm_campaign=Badge_Coverage) [![Documentation Status](https://readthedocs.org/projects/tibanna/badge/?version=latest)](https://tibanna.readthedocs.io/en/latest/?badge=latest)\n\n***\n\nTibanna runs portable pipelines (in CWL/WDL/Snakemake/shell) on the AWS Cloud.\n\n<br>\n\nInstall Tibanna.\n```bash\npip install tibanna\n```\n\n<br>\n\nUse CLI to set up the cloud component and run workflow.\n```bash\n# Deploy Unicorn to the Cloud (Unicorn = serverless scheduler/resource allocator).\ntibanna deploy_unicorn --usergroup=mygroup\n\n# Run CWL/WDL workflow on the Cloud.\ntibanna run_workflow --input-json=myrun.json\n```\n\n<br>\n\nAlternatively, use Python API.\n\n```python\nfrom tibanna.core import API\n\n# Deploy Unicorn to the Cloud.\nAPI().deploy_unicorn(usergroup='mygroup')\n\n# Run CWL/WDL workflow on the Cloud.\nAPI().run_workflow(input_json='myrun.json')\n```\n\n<br>\n\n---\nNote: Starting `0.8.2`, Tibanna supports local CWL/WDL files as well as shell commands and Snakemake workflows.\n\nNote 2: As of Tibanna version `2.0.0`, Python 3.6 is no longer supported. Please switch to Python 3.8! Python 3.7 is also supported as a fallback, but please prefer 3.8 if you can.\n\nNote 3: Starting `0.8.0`, one no longer needs to `git clone` the Tibanna repo. \n* Please switch from `invoke <command>` to `tibanna <command>`! \n* We also renovated the Python API as an inheritable class to allow development around tibanna.\n\n\nFor more details, see Tibanna [**Documentation**](http://tibanna.readthedocs.io/en/latest).\n* Also check out our [**paper in _Bioinformatics_**](https://doi.org/10.1093/bioinformatics/btz379).\n* A preprint can also be found on [**biorxiv**](https://www.biorxiv.org/content/10.1101/440974v3).\n\n",
    'author': '4DN-DCIC Team',
    'author_email': 'support@4dnucleome.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/4dn-dcic/tibanna',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.9',
}


setup(**setup_kwargs)



from tibanna.update_cost import update_cost
from tibanna.vars import AWS_REGION


config = {
    'function_name': 'update_cost_awsem',
    'function_module': 'service',
    'function_handler': 'handler',
    'handler': 'service.handler',
    'region': AWS_REGION,
    'runtime': 'python3.8',
    'role': 'tibanna_lambda_init_role',
    'description': 'update costs of a workflow run',
    'timeout': 300,
    'memory_size': 256
}


def handler(event, context):
    return update_cost(event)



# -*- coding: utf-8 -*-
import boto3
import json
import copy
from . import create_logger
from .cw_utils import TibannaResource
from datetime import datetime, timedelta
from dateutil.tz import tzutc
from .utils import (
    does_key_exist,
    read_s3,
    put_object_s3
)
from .awsem import (
    AwsemPostRunJson
)
from .exceptions import (
    StillRunningException,
    EC2StartingException,
    AWSEMJobErrorException,
    EC2UnintendedTerminationException,
    EC2IdleException,
    MetricRetrievalException,
    JobAbortedException,
    AWSEMErrorHandler
)
from .vars import PARSE_AWSEM_TIME, AWSEM_TIME_STAMP_FORMAT, S3_ENCRYT_KEY_ID
from .core import API


RESPONSE_JSON_CONTENT_INCLUSION_LIMIT = 30000  # strictly it is 32,768 but just to be safe.


logger = create_logger(__name__)


def check_task(input_json):
    return CheckTask(input_json).run()


class CheckTask(object):
    TibannaResource = TibannaResource
    API = API

    def __init__(self, input_json):
        self.input_json = copy.deepcopy(input_json)

    def run(self):
        # s3 bucket that stores the output
        bucket_name = self.input_json['config']['log_bucket']
        instance_id = self.input_json['config'].get('instance_id', '')

        # info about the jobby job
        jobid = self.input_json['jobid']
        job_started = "%s.job_started" % jobid
        job_success = "%s.success" % jobid
        job_error = "%s.error" % jobid
        job_aborted = "%s.aborted" % jobid

        public_postrun_json = self.input_json['config'].get('public_postrun_json', False)

        # check to see ensure this job has started else fail
        if not does_key_exist(bucket_name, job_started):
            start_time = PARSE_AWSEM_TIME(self.input_json['config']['start_time'])
            now = datetime.now(tzutc())
            # terminate the instance if EC2 is not booting for more than 10 min.
            if start_time + timedelta(minutes=10) < now:
                try:
                    boto3.client('ec2').terminate_instances(InstanceIds=[instance_id])
                    self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json)
                except:
                    pass  # most likely already terminated or never initiated
                raise EC2IdleException("Failed to find jobid %s, ec2 is not initializing for too long. Terminating the instance." % jobid)
            raise EC2StartingException("Failed to find jobid %s, ec2 is probably still booting" % jobid)

        # check to see if job has been aborted (by user or admin)
        if does_key_exist(bucket_name, job_aborted):
            try:
                self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json)
            except Exception as e:
                logger.warning("error occurred while handling postrun json but continuing. %s" % str(e))
            raise JobAbortedException("job aborted")

        # check to see if job has error, report if so
        if does_key_exist(bucket_name, job_error):
            try:
                self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json)
            except Exception as e:
                logger.warning("error occurred while handling postrun json but continuing. %s" % str(e))
            eh = AWSEMErrorHandler()
            if 'custom_errors' in self.input_json['args']:
                eh.add_custom_errors(self.input_json['args']['custom_errors'])
            log = self.API().log(job_id=jobid, logbucket=bucket_name)
            ex = eh.parse_log(log)
            if ex:
                msg_aug = str(ex) + ". For more info - " + eh.general_awsem_check_log_msg(jobid)
                raise AWSEMJobErrorException(msg_aug)
            else:
                raise AWSEMJobErrorException(eh.general_awsem_error_msg(jobid))

        # check to see if job has completed
        if does_key_exist(bucket_name, job_success):
            self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json)
            print("completed successfully")
            return self.input_json

        # checking if instance is terminated for no reason
        if instance_id:  # skip test for instance_id by not giving it to self.input_json
            try:
                res = boto3.client('ec2').describe_instances(InstanceIds=[instance_id])
            except Exception as e:
                if 'InvalidInstanceID.NotFound' in str(e):
                    self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json) # We need to record the end time
                    raise EC2UnintendedTerminationException("EC2 is no longer found for job %s - please rerun." % jobid)
                else:
                    raise e
            if not res['Reservations']:
                self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json) # We need to record the end time
                raise EC2UnintendedTerminationException("EC2 is no longer found for job %s - please rerun." % jobid)
            else:
                ec2_state = res['Reservations'][0]['Instances'][0]['State']['Name']
                if ec2_state in ['stopped', 'shutting-down', 'terminated']:
                    errmsg = "EC2 is terminated unintendedly for job %s - please rerun." % jobid
                    logger.error(errmsg)
                    self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json) # We need to record the end time
                    raise EC2UnintendedTerminationException(errmsg)

            # check CPU utilization for the past hour
            filesystem = '/dev/nvme1n1'  # doesn't matter for cpu utilization
            end = datetime.now(tzutc())
            start = end - timedelta(hours=1)
            jobstart_time = boto3.client('s3').get_object(Bucket=bucket_name, Key=job_started).get('LastModified')
            if jobstart_time + timedelta(hours=1) < end:
                try:
                    cw_res = self.TibannaResource(instance_id, filesystem, start, end).as_dict()
                except Exception as e:
                    raise MetricRetrievalException(e)
                if 'max_cpu_utilization_percent' in cw_res:
                    self.terminate_idle_instance(jobid,
                                                 instance_id,
                                                 cw_res['max_cpu_utilization_percent'],
                                                 cw_res['max_ebs_read_bytes'])
        # if none of the above
        raise StillRunningException("job %s still running" % jobid)

    def terminate_idle_instance(self, jobid, instance_id, cpu, ebs_read):
        if not cpu or cpu < 1.0:
            # the instance wasn't terminated - otherwise it would have been captured in the previous error.
            if not ebs_read or ebs_read < 1000:  # minimum 1kb
                # in case the instance is copying files using <1% cpu for more than 1hr, do not terminate it.
                try:
                    bucket_name = self.input_json['config']['log_bucket']
                    public_postrun_json = self.input_json['config'].get('public_postrun_json', False)
                    self.handle_postrun_json(bucket_name, jobid, self.input_json, public_read=public_postrun_json) # We need to record the end time
                    boto3.client('ec2').terminate_instances(InstanceIds=[instance_id])
                    errmsg = (
                        "Nothing has been running for the past hour for job %s,"
                        "(CPU utilization %s and EBS read %s bytes)."
                    ) %  (jobid, str(cpu), str(ebs_read))
                    raise EC2IdleException(errmsg)
                except Exception as e:
                    errmsg = (
                        "Nothing has been running for the past hour for job %s,"
                        "but cannot terminate the instance - cpu utilization (%s) : %s"
                    ) %  (jobid, str(cpu), str(e))
                    logger.error(errmsg)
                    raise EC2IdleException(errmsg)

    def handle_postrun_json(self, bucket_name, jobid, input_json, public_read=False):
        postrunjson = "%s.postrun.json" % jobid
        if not does_key_exist(bucket_name, postrunjson):
            postrunjson_location = "https://s3.amazonaws.com/%s/%s" % (bucket_name, postrunjson)
            raise Exception("Postrun json not found at %s" % postrunjson_location)
        postrunjsoncontent = json.loads(read_s3(bucket_name, postrunjson))
        prj = AwsemPostRunJson(**postrunjsoncontent)
        prj.Job.update(instance_id=input_json['config'].get('instance_id', ''))
        prj.Job.update(end_time=datetime.now(tzutc()).strftime(AWSEM_TIME_STAMP_FORMAT))
        self.handle_metrics(prj)
        logger.debug("inside funtion handle_postrun_json")
        logger.debug("content=\n" + json.dumps(prj.as_dict(), indent=4))
        # upload postrun json file back to s3
        acl = 'public-read' if public_read else 'private'
        try:
            put_object_s3(json.dumps(prj.as_dict(), indent=4), postrunjson, bucket_name,
                          public=True if acl == 'public-read' else False,
                          encrypt_s3_upload=True if S3_ENCRYT_KEY_ID else False,
                          kms_key_id=S3_ENCRYT_KEY_ID)  # defaults to None
        except Exception:  # try again no matter what
            put_object_s3(json.dumps(prj.as_dict(), indent=4), postrunjson, bucket_name,
                          public=False,  # fallback to private
                          encrypt_s3_upload=True if S3_ENCRYT_KEY_ID else False,
                          kms_key_id=S3_ENCRYT_KEY_ID)
        except Exception as e:  # noQA - catch anything if thrown from the second put call
            raise "error in updating postrunjson %s" % str(e)
        # add postrun json to the input json
        self.add_postrun_json(prj, input_json, RESPONSE_JSON_CONTENT_INCLUSION_LIMIT)

    def add_postrun_json(self, prj, input_json, limit):
        prjd = prj.as_dict()
        if len(str(prjd)) + len(str(input_json)) < limit:
            input_json['postrunjson'] = prjd
        else:
            if 'commands' in prjd:
                del prjd['commands']
            if len(str(prjd)) + len(str(input_json)) < limit:
                prjd['log'] = 'postrun json not included due to data size limit'
                input_json['postrunjson'] = prjd
            else:
                input_json['postrunjson'] = {'log': 'postrun json not included due to data size limit'}

    def handle_metrics(self, prj):
        try:
            resources = self.TibannaResource(prj.Job.instance_id,
                                             prj.Job.filesystem,
                                             prj.Job.start_time_as_datetime,
                                             prj.Job.end_time_as_datetime)

        except Exception as e:
            raise MetricRetrievalException("error getting metrics: %s" % str(e))
        prj.Job.update(Metrics=resources.as_dict())
        self.API().plot_metrics(prj.Job.JOBID, directory='/tmp/tibanna_metrics/',
                           force_upload=True, open_browser=False,
                           endtime=prj.Job.end_time_as_datetime,
                           filesystem=prj.Job.filesystem,
                           instance_id=prj.Job.instance_id)
